var fs = require("fs");
var xml2js = require("xml2js");
var parser = new xml2js.Parser();
var path = require('path');
var bundleJson = require('../bundle.json');

const MANIFEST_PATH = "./.repo/manifests/default.xml";
const COMPONENTS_INFO_PATH = "./build/lite/components";
const PRODUCT_PATH = "./vendor/hisilicon";

/**
 * 解析repo的xml文件，获取所有的项目
 * @returns 项目列表{name,path}
 */
async function parseRepoXML({ TRACE_MANIFEST_PATH = MANIFEST_PATH }) {
  if (!fs.existsSync(TRACE_MANIFEST_PATH)) {
    console.log(`${TRACE_MANIFEST_PATH} is not exist!`);
  }
  var content = fs.readFileSync(TRACE_MANIFEST_PATH);
  var obj = await parser.parseStringPromise(content);
  var projects = obj.manifest.project.map(p => ({ name: p["$"].name, path: p["$"].path }));
  return projects;
}

/**
 * 解析build/lite/components文件夹下的所有子系统json数据
 * @returns 子系统
 */
function getSubSystems({ TRACE_COMPONENTS_INFO_PATH = COMPONENTS_INFO_PATH }) {
  const dirs = fs.readdirSync(TRACE_COMPONENTS_INFO_PATH);
  const subsystems = [];
  dirs.forEach(d => {
    if (d.endsWith(".json")) {
      const subsystem = {
        subsystem: d,
        ...JSON.parse(fs.readFileSync(path.join(TRACE_COMPONENTS_INFO_PATH, d)))
      };
      subsystems.push(subsystem);
    }
  })
  return subsystems;
}

function getAllProducts({ TRACE_PRODUCT_PATH = PRODUCT_PATH }) {
  const dirs = fs.readdirSync(TRACE_PRODUCT_PATH, { withFileTypes: true })
  return dirs
    .filter(file => file.isDirectory() && file.name !== '.gitee')
    .map(pro => {
      const proPath = path.resolve(TRACE_PRODUCT_PATH, pro.name, 'config.json')
      const config = JSON.parse(fs.readFileSync(proPath));
      return {
        name: pro.name,
        kernel: config.kernel_type,
        board: config.board
      }
    })
}

/**
 * 获取产品的直接依赖（config.json中定义的）
 * @param {*} product 产品名称
 * @returns 依赖的组件
 */
function getDirectDeps(product, { TRACE_PRODUCT_PATH = PRODUCT_PATH }) {
  const configPath = path.join(TRACE_PRODUCT_PATH, product, "config.json");
  const config = JSON.parse(fs.readFileSync(configPath));
  let deps = [];
  config.subsystems.forEach(subsystem => {
    subsystem.components.forEach(c => {
      deps.push(c.component);
    })
  })
  return deps;
}

function getAllDeps({ name: product, kernel, board }, subsystems, config) {
  let allComponents = [];
  subsystems.forEach(subsystem => allComponents = allComponents.concat(subsystem.components));

  // 直接依赖
  const deps = getDirectDeps(product, config);

  const allDeps = [];

  deps.forEach(dep => {
    let entry = [dep];
    while (entry.length > 0) {
      const current = entry.pop();
      let currentDeps = [];
      if (!allDeps.includes(current)) {
        allDeps.push(current);
        const component = allComponents.find(c => c.component === current);
        const comDeps = component && component.deps;
        if (comDeps) {
          if (comDeps["third_party"]) {
            currentDeps = currentDeps.concat(comDeps["third_party"].map(c => c.toLowerCase().replace(/-/g, "_")));
          }
          if (comDeps["components"]) {
            currentDeps = currentDeps.concat(comDeps["components"]);
          }
          if (comDeps["kernel_special"] && comDeps["kernel_special"][kernel]) {
            currentDeps = currentDeps.concat(comDeps["kernel_special"][kernel]);
          }
          if (comDeps["board_special"] && comDeps["board_special"][board]) {
            currentDeps = currentDeps.concat(comDeps["board_special"][board]);
          }
          entry = entry.concat(currentDeps);
        }
      }
    }
  })

  return allDeps;
}

function outPutJsonFormate(result = {}, { json = false, object = 'product' }) {
  if (json) {
    const res = {
      [`${object}s`]: result
    }
    return JSON.stringify(result, null, 2);
  } else {
    // 处理object对象
    let res = [];
    result.children.forEach(r => {
      res.push(r.name);
      r.children && r.children.forEach(rc => {
        res.push(rc.name);
        rc.children && rc.children.forEach(rcc => {
          res.push(rcc.name);
        })
      })
    })
    return [...new Set(res)].join('\n');
  }
}

// 将结果data写入.output目录下data.json
function writeDataTree(hpmApi, data) {
  const outputDir = hpmApi.getOutputDir(bundleJson.name);
  hpmApi.shelljs.cp('-r', path.resolve(__dirname, '..', 'ui'), outputDir);
  fs.writeFileSync(path.join(outputDir, "ui", 'data.json'), JSON.stringify(data, null, 4));
}

/**
 * 获取多级节点组件依赖
 * @param 根节点组件树：nodeObj 命令参数(层级)：opts 子系统：sys
 * @returns 组件依赖树
 */
function getNatchComponents(nodeObj,depth,sys) {
  if (!nodeObj) {
    return false;
  }
  let T = getRelyFirst(nodeObj.name, sys),num = depth;
  if (num > 3) {
    // 设置最深三层依赖
    num = 3;
  }
  T && T.forEach(t => {
    nodeObj.children.push(t);
  })
  if (num > 1) {
    nodeObj.children && nodeObj.children.forEach(dc => {
      getNatchComponents(dc, num-1, sys);
    })
  }
  return nodeObj;
}

/**
 * 构建根节点组件直接依赖树
 * @param 根节点名称：rootName，子系统：subSystems
 * @returns 直接依赖数组
 */
function getRelyFirst(rootName, subSystems) {
  if (rootName === '') {
    return [];
  }
  let _objs = []
  subSystems.forEach(s => {
    const components = s.components;
    components.forEach(c => {
      c.deps.components && c.deps.components.concat(c.deps.third_party).forEach(cdcs => {
        if(rootName === (cdcs && cdcs.toLowerCase().replace(/-/g, "_"))) {
          // dirs为数组
          _objs.push({name: c.component, children: []});
        }
      })
    })
  })
  return _objs
}

/**
 * 遍历树节点，存在闭环依赖，重新构建树结构
 * @param {*} T 原始的树结构
 * @returns 依赖组件树：标记闭环依赖节点并去除该节点的子节点
 */
function checkNodeSame(t) {
  t.children && t.children.forEach(tc => {
    checkNodeSame(tc)
    tc.children && tc.children.forEach(tcc => {
      checkNodeSame(tcc)
      if (t.name === tcc.name) {
        tcc.label = null;
        tcc.label = {color: 'red'};
        tcc.itemStyle = {color: 'red'};
        tcc.children = [];
      }
    })
  })
  return t;
}

async function start(options, args, hpmApi) {
  // 根节点名称
  let rootName = '';
  // 是否为直接组件
  let componentCheck = false;
  const repoNames = args[0];
  const config = hpmApi.getConfig()
  const subSystems = getSubSystems(config);
  const repos = await parseRepoXML(config);
  const changedRepos = repos.filter(p => repoNames.includes(p.name));
  // 查询project名称关联组件不存在时，继续查询依赖的组件名称（重组关联组件）
  // 查询project名称存在（普通project或者第三方），解析至关联组件
  if(changedRepos.length == 0) {
    componentCheck = true;
    rootName = repoNames;
  }
  if (changedRepos.length > 0 || componentCheck) {
    if (componentCheck) {} else {
      const paths = changedRepos.map(p => p.path);
      // 遍历所有的子系统的组件清单，筛选出包含了该路径（前缀）的组件名，获取一级依赖
      // 判断如果为第三方组件，解析组件名，获取一级依赖
      if (paths.filter(p => p.startsWith("third_party")).length != 0) {
        const thirdPartyComponents = paths.filter(p => p.startsWith("third_party"))
          .map(t => {
            return t.substr("third_party/".length)
              .toLowerCase()
              .replace(/-/g, "_");
          })
        rootName = thirdPartyComponents[0];
      } else {
        // 正常搜索project名称
        const projectComponents = []
        subSystems.forEach(s => {
          const components = s.components;
          components.forEach(c => {
            if (c.dirs && c.dirs.find(dir => paths.some(p => dir.startsWith(p)))) {
              projectComponents.push(c.component);
            }
          })
        })
        rootName = projectComponents[0]
      }
    }
    
    // 处理多层构建树，参数：根节点对象、层级参数、子系统
    let matchTrees = getNatchComponents({name: rootName, children: []}, Number(options.depth), subSystems);

    // 处理最终结果依赖树，存在闭环节点做提示并不展示其后面的子节点
    checkNodeSame(matchTrees);

    // 上述组件任意一个被包含在产品的组件清单中，则返回该产品
    let matchedComponents = outPutJsonFormate(matchTrees, {json: false}).split('\n');
    const matchedProducts = [];
    const products = getAllProducts(config)
    products.forEach(product => {
      const allDeps = getAllDeps(product, subSystems, config);
      if (matchedComponents.filter(c => allDeps.includes(c)).length > 0) {
        matchedProducts.push({name: product.name});
      }
    })
    if ( options.object === 'bundle' ) {
      // 展示树返回
      writeDataTree(hpmApi, [matchTrees])
      // 命令行返回
      return outPutJsonFormate(matchTrees, options)
    } else {
      // 产品影响展示树返回
      writeDataTree(hpmApi, [{name: rootName, children: matchedProducts}])
    }
    return outPutJsonFormate({name: rootName, children: matchedProducts}, options);
  } else {
    throw `Cannot found repo by ${repoNames}`;
  }
}

module.exports = { start }